/*
 * this dumps FASTA sequence and QUALITIY values from SFF File(s)
 *
 * syntax:
 * sff_dump -f EIK123.fna -q EIK123.qual EIK123.sff
 *
 * Bernd Senf - bsenf@fli-leibniz.de, 2007-03-29
 * with suggestions from James Bonfield (jbk)
 *
 * Some optimazation hints from flowers at shgc.stanford.edu
 * till to 33% faster - 2008-10-24
 */


#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include "sff.h"
#include "os.h"
#include "mFILE.h"
#include "misc.h"    /* Min, Max */


typedef struct  
{
 	char  trace_name[32];
 	int   NBases;
 	int   qual_left;
 	int   qual_right;
 	char  bases[8192];
 	int   quality[8192];
    int   leftCutoff;
    int   rightCutoff;
 
} sff_dump_struct;


static sff_dump_struct SFF;


void reverse(char s[])
{
  int i, j;
  char c;

  for (i = 0, j = strlen(s) - 1; i < j; i++, j--)
  {
    c = s[i];
    s[i] = s[j];
    s[j] = c;
  }
}

char* itoa(int n, char s[])
{
  int i;

  i = 0;
  do
  {
     s[i++] = n % 10 + '0';
  }
  while ((n /= 10) > 0);
  s[i] = '\0';
  reverse(s);

  return s;
}



/*
 * Reads an SFF file from an mFILE and decodes it to a static SFF structure
 *
 * Returns <>NULL on success
 *         NULL on failure
 */
int mfread_sff_recurive(mFILE *mf, int flow_len,char* flow_order) 
{

	int i;
	int nbases;
	sff_read_header *rh;
	sff_read_data *rd;

 	/* simple and fast and secure 
       No need of extra test of trailing \0 for strings : */
 
	memset(&SFF,0,sizeof(SFF));
 
	if (NULL == (rh = read_sff_read_header(mf))) 
	{
    	return(0);
	}
	
	if (NULL == (rd = read_sff_read_data(mf, flow_len, rh->nbases))) 
	{
		free_sff_read_header(rh);
    	return (0);
	}
 
	strncpy(SFF.trace_name,rh->name,rh->name_len);
 
	SFF.NBases = nbases = rh->nbases;
 
	strncpy(SFF.bases,rd->bases,nbases);
 
	for (i=0;i<SFF.NBases;i++) 
	{
		SFF.quality[i]=rd->quality[i];
	}

	SFF.leftCutoff = MAX(rh->clip_qual_left, rh->clip_adapter_left);

	SFF.rightCutoff = MIN(rh->clip_qual_right ? rh->clip_qual_right : nbases+1,
                  rh->clip_adapter_right ? rh->clip_adapter_right : nbases+1);
 
	free_sff_read_header(rh);
	free_sff_read_data(rd);

	return (1);
}


int main(int argc, char **argv) 
{
    int i;
    sff_common_header *ch;
 
    char *sff=NULL;
 	char *qual=NULL;
 	char *fna=NULL;

	int iqual=0;
	int ifna=0;
	
 	FILE *fp_qual=NULL;
	FILE *fp_fna=NULL;
 	FILE *fp_sff=NULL;
 
    for (argc--, argv++; argc > 0; argc--, argv++) 
 	{
  		if (strcmp(*argv, "-q") == 0) 
  		{
      		qual = *++argv;
      		argc--;
     	} 
  		else if (strcmp(*argv, "-f") == 0) 
  		{
      		fna = *++argv;
      		argc--;
     	} 
  		else 
  		{
    		sff=*argv;
  		}
 	}
 
 	if ( argc > 0 || sff == NULL || ( !qual && !fna ) ) 
	{
  		printf("Usage: sff_dump [-h] [-f dump.fna ] [-q dump.qual ] trace.sff\n");
  		exit(1);
 	}
 
 	if (fna &&  (NULL==(fp_fna=fopen(fna,"w")))) 
	{
  		fprintf(stderr,"Can't write to FNA (%s)",fna);
  		perror(" ");
  		exit (1);
 	}
	
 	if (qual &&  (NULL==(fp_qual=fopen(qual,"w")))) 
 	{
  		fprintf(stderr,"Can't write to QUAL (%s)",qual);
  		perror(" ");
  		exit (1);
 	}

 	if (sff && (NULL==(fp_sff=fopen(sff,"rb")))) 
 	{ 
  		fprintf(stderr,"Can't read SFF (%s)",sff);
  		perror(" ");
		exit(1);
 	} 
	else 
	{
		int magic;
		fread(&magic,4,1,fp_sff);
		
		magic=be_int4(magic);
		fclose(fp_sff); 
		if (magic != SFF_MAGIC )
		{
			fprintf(stderr,"Ops, non correct SFF magic [%4.4X] [%s]\n",magic,sff);
			exit(1); 
		}
	}
 
 
 	/* ------- parameter checks finished --------------- */
 
	mFILE *mf = mfopen(sff, "rb");
 
    if (NULL == (ch = read_sff_common_header(mf))) 
 	{
  		free_sff_common_header(ch);
  		exit(1);
 	}
 
 	i=0; 
 	while (mfread_sff_recurive(mf,ch->flow_len,ch->flow))
 	{
    
  		if (fna && fp_fna )  
  		{
    		fprintf(fp_fna,">%s LEN=%d QL=%d QR=%d\n",SFF.trace_name,SFF.NBases,SFF.leftCutoff,SFF.rightCutoff);
			fflush(fp_fna);
    		fprintf(fp_fna,"%s\n",SFF.bases);
			fflush(fp_fna);
			ifna++;  	/* statistic */
  		}
  
  		if (qual && fp_qual) 
  		{
    		char szQuality[8192];
    		int j;
			char szTemp[16];
    
    		memset(szQuality,0,sizeof(szQuality));
    
    		fprintf(fp_qual,">%s LEN=%d QL=%d QR=%d\n",SFF.trace_name,SFF.NBases,SFF.leftCutoff,SFF.rightCutoff);
			fflush(fp_qual);
    		for (j=0;j<SFF.NBases;j++)
			{
     			strcat(szQuality,itoa(SFF.quality[j],szTemp));
     			if (j <  SFF.NBases-1 ) strcat(szQuality," ");
    		}
    		fprintf(fp_qual,"%s\n",szQuality); 
			fflush(fp_qual);
			iqual++;
  		}
		
    	/* Pad to 8 chars */
    	mfseek(mf, (mftell(mf) + 7)& ~7, SEEK_SET);
 	}
 
 	free_sff_common_header(ch);
 	mfclose(mf);
	
	if (fp_qual) fclose(fp_qual);
	if (fp_fna)  fclose(fp_fna);
	
	fprintf(stderr,"%s: %d entries written.\n",fna,ifna);
	fprintf(stderr,"%s: %d entries written.\n",qual,iqual);
	
 	exit(0);
}
